#!/usr/bin/perl -w

#   Copyright (C) 2006 David G. Jacobowitz 
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#

package dataqual;

use warnings qw(all);
use strict 'refs';

use Exporter qw(import);
use POSIX qw(floor);
use eqn_string qw(create_evaluable_eqn);
use misc2 qw(func_fail);
use Data::Dumper;

use Scalar::Util qw(looks_like_number);

@EXPORT_OK = qw(gen_defquald_aa gen_defquald_ha gen_defquald_a
                gen_numquald_aa gen_numquald_ha gen_numquald_a
		gen_ifquald_aa  gen_ifquald_ha  gen_ifquald_a);

# applies if array, and filters out both undefined 
# and non-numeric values
sub gen_numquald_aa {
 my $dm    = shift;
 my $vlist = shift;
 my $ifary = shift;

 my @vars = @{$vlist};
 if (@vars == 0) {
  return undef;
 }
 my $firstvar = $vars[0];
 my $depth = @{$dm->{data}{$firstvar}};
 my $q_aa = [];
 my $ct = 0;
 for (my $row=0;$row<$depth;$row++) {
  my $defd = 1;
  for (my $col=0;$col<@vars;$col++) {
#  print "$col $row VAL " . $dm->{data}{$vars[$col]}[$row] . "\n";
   if (!looks_like_number($dm->{data}{$vars[$col]}[$row])) {
    $defd = 0;
   }
#  print "defined looks_like_number $defd\n";
  }
  my $quald = $defd && (!defined($ifary) || ($ifary->[$row]));
  if ($quald) {
   for (my $col=0;$col<@vars;$col++) {
    $q_aa->[$col][$ct] = $dm->{data}{$vars[$col]}[$row];
   }
   $ct++;
  }
 }
 return $q_aa;
}


sub gen_numquald_a {
 my $dm    = shift;
 my $v     = shift;
 my $ifary = shift;
 
 my $depth = @{$dm->{data}{$v}};
 my $a     = [];
 my $ct    = 0;
 for (my $row=0;$row<$depth;$row++) {
  my $numd  = looks_like_number($dm->{data}{$v}[$row]);
  my $quald = $numd && (!defined($ifary) || ($ifary->[$row]));
  if ($quald) {
   $a->[$ct++] = $dm->{data}{$v}[$row];
  }
 }
 return $a;
}



# applies if array and filters out undfined values
sub gen_defquald_aa {
 my $dm    = shift;
 my $vlist = shift;
 my $ifary = shift;

 my @vars = @{$vlist};
 if (@vars == 0) {
  return undef;
 }
 my $firstvar = $vars[0];
 my $depth = @{$dm->{data}{$firstvar}};
 my $q_aa = [];
 my $ct = 0;
 for (my $row=0;$row<$depth;$row++) {
  my $defd = 1;
  for (my $col=0;$col<@vars;$col++) {
  #print "VAL " . $dm->{data}{$vars[$col]}[$row] . "\n";
   if ($dm->{data}{$vars[$col]}[$row] eq "__undef") {
    $defd = 0;
   }
  }
  my $quald = $defd && (!defined($ifary) || ($ifary->[$row]));
  if ($quald) {
   for (my $col=0;$col<@vars;$col++) {
    $q_aa->[$col][$ct] = $dm->{data}{$vars[$col]}[$row];
   }
   $ct++;
  }
 }
 return $q_aa;
}


sub gen_defquald_a {
 my $dm    = shift;
 my $v     = shift;
 my $ifary = shift;
 my $depth = @{$dm->{data}{$v}};
 my $a     = [];
 my $ct    = 0;
 for (my $row=0;$row<$depth;$row++) {
  my $defd = (defined($dm->{data}{$v}[$row]) &&
	      ($dm->{data}{$v}[$row] ne "__undef"));

  my $quald = $defd && (!defined($ifary) || ($ifary->[$row]));
  if ($quald) {
   $a->[$ct++] = $dm->{data}{$v}[$row];
  }
 } 
 return $a;
}


### only applies "if" array; undefined values *are* inluded -- but if 
### any item in the row was "undef" they will all be.... I'm not sure 
### if this is what I want...
sub gen_ifquald_aa {
 my $dm    = shift;
 my $vlist = shift;
 my $ifary = shift;

 my @vars = @{$vlist};
 if (@vars == 0) {
  return undef;
 }
 my $firstvar = $vars[0];
 my $depth = @{$dm->{data}{$firstvar}};
 my $q_aa = [];
 my $ct = 0;
 for (my $row=0;$row<$depth;$row++) {
  my $defd = 1;
  for (my $col=0;$col<@vars;$col++) {
  #print "VAL " . $dm->{data}{$vars[$col]}[$row] . "\n";
   if ($dm->{data}{$vars[$col]}[$row] eq "__undef") {
    $defd = 0;
   }
  }
  my $quald = (!defined($ifary) || ($ifary->[$row]));
  if ($quald) {
   for (my $col=0;$col<@vars;$col++) {
    if ($defd) {
     $q_aa->[$col][$ct] = $dm->{data}{$vars[$col]}[$row];
    } else {
     $q_aa->[$col][$ct] = "__undef";
    }
   }
   $ct++;
  }
 }
 return $q_aa;
}


# like ifquald_aa but only generates a single array ref for one
# variable -- this is what you want to call, even repeatedly if you
# don't want the qualification "anded" across variables.
sub gen_ifquald_a {
 my $dm    = shift;
 my $v     = shift;
 my $ifary = shift;

 my $a = [];

 my $depth = @{$dm->{data}{$v}};
 my $ct = 0;
 if (defined($ifary)) {
  for (my $row=0;$row<$depth;$row++) {
   if ($ifary->[$row]) {
     $a->[$ct++] = $dm->{data}{$v}[$row];
   }
  }
 } else {
  $a = $dm->{data}{$v};
 }
 return $a;
}


sub gen_numquald_ha {
 my $dm    = shift;
 my $vlist = shift;
 my $ifary = shift;

 my @vars = @{$vlist};
 if (@vars == 0) {
  return undef;
 }
 my $firstvar = $vars[0];
 my $depth = @{$dm->{data}{$firstvar}};
 my $q_ha = {};
 $q_ha->{orig_depth} = $depth; 
 my $ct = 0;
 for (my $row=0;$row<$depth;$row++) {
  my $defd = 1;
  foreach my $var (@vars) {
   if (!looks_like_number($dm->{data}{$var}[$row])) {
    $defd = 0;
   }
  }
  my $quald = $defd && (!defined($ifary) || ($ifary->[$row]));
  if ($quald) {
#   print "quald at row $row\n";
   foreach my $var (@vars) {
    $q_ha->{data}{$var}[$ct] = $dm->{data}{$var}[$row];
   }
   $q_ha->{orig_from}[$ct] = $row;
#   print "CT $ct row $row\n";
   $ct++;
  }
 }
 foreach my $var (@vars) {
  $q_ha->{info}{$var}{type} = "numeric";
 } 
 return $q_ha;
}

sub gen_defquald_ha {
 my $dm    = shift;
 my $vlist = shift;
 my $ifary = shift;

 my @vars = @{$vlist};
 if (@vars == 0) {
  return undef;
 }
 my $firstvar = $vars[0];
 my $depth = @{$dm->{data}{$firstvar}};
 my $q_ha = {};
 my $ct = 0;
 my $nums = {};
 foreach my $var (@vars) {
  $nums->{$var} = 1;
 }
 $q_ha->{orig_depth} = $depth; 
 for (my $row=0;$row<$depth;$row++) {
  my $defd = 1;
  foreach my $var (@vars) {
   if (!defined($dm->{data}{$var}[$row]) ||
       ($dm->{data}{$var}[$row] eq "__undef")) {
    $defd = 0;
   }
  }
  my $quald = $defd && (!defined($ifary) || ($ifary->[$row]));
  if ($quald) {
   foreach my $var (@vars) {
    if (!looks_like_number($dm->{data}{$var}[$row])) {
     $nums->{$var} = 0;
    }
    $q_ha->{data}{$var}[$ct] = $dm->{data}{$var}[$row];
   }
   $q_ha->{orig_from}[$ct] = $row;
   $ct++;
  }
 }
 foreach my $var (@vars) {
  $q_ha->{info}{$var}{type} = ($nums->{$var}) ? "numeric" : "categorical";
 }
 return $q_ha;
}


sub gen_ifquald_ha {
 my $dm    = shift;
 my $vlist = shift;
 my $ifary = shift;

 my @vars = @{$vlist};
 if (@vars == 0) {
  return undef;
 }
 my $firstvar = $vars[0];
 my $depth = @{$dm->{data}{$firstvar}};
 my $q_ha = {};
 my $ct = 0;
 my $nums = {};
 foreach my $var (@vars) {
  $nums->{$var} = 1;
 }
 $q_ha->{orig_depth} = $depth; 
 for (my $row=0;$row<$depth;$row++) {
  my $defd = 1;
  foreach my $var (@vars) {
   if (!defined($dm->{data}{$var}[$row]) ||
       ($dm->{data}{$var}[$row] eq "__undef")) {
    $defd = 0;
   }
  }
  my $quald = (!defined($ifary) || ($ifary->[$row]));
  if ($quald) {
   foreach my $var (@vars) {
    if ($defd) { 
     $q_ha->{data}{$var}[$ct] = $dm->{data}{$var}[$row];
    } else {
     $q_ha->{data}{$var}[$ct] = "__undef";
    }
    if ($defd && (!looks_like_number($dm->{data}{$var}[$row]))) {
     $nums->{$var} = 0;
    }
   }
   $q_ha->{orig_from}[$ct] = $row;
   $ct++;
  }
 }
 foreach my $var (@vars) {
  $q_ha->{info}{$var}{type} = ($nums->{$var}) ? "numeric" : "categorical";
 }
 return $q_ha;
}


1;
