#!/usr/bin/perl -w

#   Copyright (C) 2006 David G. Jacobowitz 
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#

use fgroup;
package fgroup::graph;

use warnings 'all';
use strict   'refs';

use Exporter 'import';
@EXPORT_OK = qw();

use Data::Dumper;
use File::Which qw(which where);
use Scalar::Util qw(looks_like_number);

use misc qw(fixedFloat fixedFloat_p one_var_stats deep_copy);
use misc2 qw(func_fail tstFlg setFlg);
use dataqual qw(gen_defquald_aa gen_numquald_aa
                gen_defquald_a  gen_numquald_a);

# use Statistics::Distributions;
use mod_distributions;

use strict;

my $PI          = 3.14159265358;
my $TWOPI       = 2 * $PI;
my $MAXPOINTS   = 200;
my $GRAPH_DEFAULT_FORMAT = "png";
my $graph_debug = 1;


sub new {
 my $proto           = shift;
 my $oo              = shift;
 my $gb              = shift;
 my $dm              = shift;
 my $self            = {};

 $self->{dm} = $dm;
 $self->{gb} = $gb;
 $self->{oo} = $oo;

 my $gnuplot = which('gnuplot');
 if (defined($gnuplot)) {
  $self->{gnuplot}      = $gnuplot;
  $self->{have_gnuplot}     = 1;
  $self->{gb}{have_gnuplot} = 1;
 } else {
  $self->{have_gnuplot}     = 0;
  $self->{gb}{have_gnuplot} = 0;
 }

 my $class           = ref($proto) || $proto;
 bless ($self,$class);
 return $self;
}

sub commands {
 my $self = shift;
 my @cmds = ("histo","histogram", "scatter");
 return @cmds;
}

sub histo {
 my $self = shift;
 my $a    = shift;
 return $self->histogram($a);
}

sub histogram {
 my $self = shift;
 my $a    = shift;

 my $r    = {};
 $r->{status} = "ok";

 if (!$self->{have_gnuplot}) {
  return { status=> "fail; GNUplot not detected on this system. Cannot generate plots" };
 }

 if (!defined($a->{varary}[0]{state}) || ($a->{varary}[0]{state} ne "exist")) {
  return { status=> "fail; no variable specified or unknown variable" };
 } 

 my $format = $GRAPH_DEFAULT_FORMAT;
 if (defined($a->{options}{format})) {
  $format = $a->{options}{format}{str};
 }

 my $oname = "$$";
 if (defined($a->{options}{output})) {
  $oname = $a->{options}{output}{str};
 }

 if ($format eq "sheet") {
  $oname .= ".csv";
 } else {
  $oname .= "." . $format;
 }

 $self->create_histo_plot($a,
		          $a->{varary}[0]{name},
	                  $format,
	                  $oname);

 return $r;
}


sub scatter {
 my $self = shift;
 my $a    = shift;
 my $r    = {};
 $r->{status} = "ok";

 if (!$self->{have_gnuplot}) {
  return { status=> "fail; GNUplot not detected. Plotting disabled." };
 }

 my @vars = ();
 foreach my $v (@{$a->{varary}}) {
  if (defined($v->{state}) && ($v->{state} eq "exist")) {
   push(@vars,$v->{name});
  } else {
   return { status => "fail; variable $v->{name} does not exist." };
  }
 }

 my $vcount = @vars;
 if ($vcount<2) {
  return { status => "fail; must provide at least two variables" };
 } elsif ($vcount>4) {
  return { status => "fail; must provide no more than three variables" };
 }

 my $format = $GRAPH_DEFAULT_FORMAT;
 if (defined($a->{options}{format})) {
  $format = $a->{options}{format}{str};
 }

 my $oname = "$$";
 if (defined($a->{options}{output})) {
  $oname = $a->{options}{output}{str};
 }

 my $gridresx = 20;
 my $gridresy = 20;

 if (looks_like_number($a->{options}{gridresx}{str})) {
  $gridresx = $a->{options}{gridresx}{str};
 }
 if (looks_like_number($a->{options}{gridresy}{str})) {
  $gridresy = $a->{options}{gridresy}{str};
 }

 if (looks_like_number($a->{options}{gridres}{str})) {
  $gridresx = $a->{options}{gridres}{str};
  $gridresy = $a->{options}{gridres}{str};
 }

 my $title = '';
 if (defined($a->{options}{title})) {
  $title = $a->{options}{title}{str};
 }


 if ($format eq "sheet") {
  $oname .= ".csv";
 } else {
  $oname .= "." . $format;
 }

 $self->create_scatter_plot($a,
		            \@vars,
	                    $format,
			    $gridresx, $gridresy,
	                    $oname,
		            $title);

 return $r;
}



sub create_scatter_plot {
 my $self    = shift;
 my $a       = shift;
 my $varrf   = shift;
 my $fmt     = shift;
 my $gridresx= shift;
 my $gridresy= shift;
 my $oname   = shift;
 my $title   = shift;

 my @vars   = @{$varrf};
 my $vcount = @vars;

 my $dm     = $self->{dm};
 my $gb     = $self->{gb};
 my $oo     = $self->{oo};

 my $qaa = gen_numquald_aa($dm, $varrf, $a->{ifary});

 my $dname = $oname . ".tmp.dat";
 open(DF, "> " . $dname) || return { status => "fail; count not open temp data file for write" };
# foreach my $v (@vars) {
#  print DF "$v\t";
# }
# print DF "\n";

 for (my $i=0;$i<@{$qaa->[0]};$i++) {
  for (my $j=0;$j<@{$qaa};$j++) {
   print DF "$qaa->[$j][$i]\t";
  }
  print DF "\n";
 }
 close DF;

 if ($fmt eq "eps") {
  $fmt = "postscript eps enhanced color";
 }

 # project implies color; make it so
 setFlg($a,"color",tstFlg($a,"color") || tstFlg($a,"project") || tstFlg($a,"map"));

 my $cname = $oname . ".tmp.cmd";
 open (CF, "> " . $cname) || return { status => "fail; could not open temp cmd file" };
 my $os = "";
 $os .= "set term $fmt size 2048,1536\n";
 $os .= "set output \"$oname\"\n";
 $os .= "unset hidden3d\n";
 $os .= "set ticslevel 0.5\n";
 $os .= "set view 60,30\n";
 $os .= "set autoscale\n";
# $os .= "set parametric\n";
 $os .= "set xlabel \'$vars[0]\'\n";
 $os .= "set ylabel \'$vars[1]\'\n";
 if (defined($title) && length($title)) {
  $os .= "set title \"$title\"\n";
 }
 if ($vcount == 2) {
  $os .= "set style data points\n";
  $os .= "plot \"$dname\" index 0 title \'$vars[0] vs. $vars[1]\'\n";
 } else {
  if (tstFlg($a,"color")) {
   if (tstFlg($a,"map")) {
    $os .= "set pm3d map\n";
    $os .= "set dgrid3d $gridresx,$gridresy\n";
   } elsif (tstFlg($a,"project")) {
    $os .= "set pm3d at b\n";
    $os .= "set dgrid3d $gridresx,$gridresy\n";
   } else {
    $os .= "set pm3d\n";
    $os .= "set dgrid3d $gridresx,$gridresy\n";
   }
  }
  $os .= "set zlabel \'$vars[2]\'\n";
  $os .= "splot \"$dname\" index 0 title\'$vars[0] vs. $vars[1] vs. $vars[2]\'\n";
 }
 $os .= "quit\n";
 print CF $os;
 close CF;

 if (1) {
  $oo->disp("Calling gnuplot...");
  my $res = `$self->{gnuplot} $cname`;
  $oo->disp(sprintf("done. %s\n",length($res)?"result: $res":""));
  if (-e $oname) {
   $oo->disp("Created plot \"$oname\"\n");
  } else {
   return { status => "fail; graph was not created. Is gnuplot installed?" }
  }

  if (!$graph_debug) {
   $oo->disp("");
   unlink $cname;
   unlink $dname;
  }
 }

 return { status => "ok" };
}







sub create_histo_plot {
 my $self  = shift;
 my $a     = shift;
 my $var   = shift;
 my $fmt   = shift;
 my $oname = shift;
 my $r = undef;

 my $dm    = $self->{dm};
 my $gb    = $self->{dm};
 my $oo    = $self->{oo};

 my $ba = [];
 my $normal = {};
 my $bincount = 10;
 my $discrete = tstFlg($a,"discrete");
 my $do_normal = tstFlg($a,"normal");
 my $some_dropped = 0;

 if ($discrete && $do_normal) {
  return { status => "fail; cannot overlay normal curve on discrete histogram" };
 }

 if ($discrete) {
  my $qa = gen_defquald_a($dm,$var,$a->{ifary});
  my $bh = {};
  for (my $i=0;$i<@{$qa};$i++) {
   if (defined($bh->{ $qa->[$i] })) {
    $bh->{ $qa->[$i] } += 1;
   } else {
    $bh->{ $qa->[$i] } = 1;
   }
  }
  my $i=0;

  sub numSortIfPossible ($$) {
   if (looks_like_number($_[0]) && looks_like_number($_[1])) {
    return ($_[0] <=> $_[1]);
   } else {
    return ($_[0] cmp $_[1]);
   }
  }

  foreach my $key (sort numSortIfPossible keys %{$bh}) {
   $ba->[$i]{name}   = $key;
   $ba->[$i]{center} = $i;
   $ba->[$i]{freq}   = $bh->{$key};
   $i++;
  }
  $bincount = $i;
 } else {

  $bincount = 10;
  if (defined($a->{options}{bincount})) {
   if (looks_like_number($a->{options}{bincount}{str})) {
    $bincount = $a->{options}{bincount}{str};
   } else {
    return { status => "fail; illegal bincount specified" };
   }
  }
  
  my $qa = gen_numquald_a($dm, $var, $a->{ifary});
  $r = one_var_stats($qa);
  if (func_fail($r)) { return $r; };
  my $min = $r->{min};
  my $max = $r->{max};
  $normal = { min       => $min,
	      max       => $max,
	      mean      => $r->{mean},
	      sd        => $r->{std_dev},
	      obs       => $r->{count},
	      bincount  => $bincount,
	      do_normal => $do_normal };


  if (defined($a->{options}{minval})) {
   if (looks_like_number($a->{options}{minval}{str})) {
    $min = $a->{options}{minval}{str};
   } else {
    return { status => "fail; illegal minval specified" };
   }
  }
  if (defined($a->{options}{maxval})) {
   if (looks_like_number($a->{options}{maxval}{str})) {
    $max = $a->{options}{maxval}{str};
   } else {
    return { status => "fail; illegal maxval specified" };
   }
  }

  my $bwidth = ($max-$min)/$bincount;
  if ($bwidth <= 0) {
   return { status => "fail; negative/zero bin width" };
  }

  for (my $bc=0;$bc<$bincount;$bc++) { 
   $ba->[$bc]{freq} = 0; 
   my $binmin = $min + $bc*$bwidth;
   my $binmax = $min + ($bc+1)*$bwidth;

   # round!
   $binmin = int($binmin*10+0.5)/10;
   $binmax = int($binmax*10+0.5)/10;

   $ba->[$bc]{name} = "$binmin<=X<$binmax";
   $ba->[$bc]{center} = $min + ($bc*$bwidth) + $bwidth/2;
  };
  for (my $i=0;$i<@{$qa};$i++) {
   my $v = $qa->[$i];
   my $bin = int(($v-$min) / $bwidth);
   if ($v == $max) { $bin = $bincount-1; };
   if (defined($ba->[$bin])) {
    $ba->[$bin]{freq} += 1;
   } else {
    $some_dropped += 1;
   }
  }
 }

 if ($some_dropped) {
  $oo->disp("Because of min/max in place for histogram, $some_dropped values have been dropped.\n");
 }

 my $cname = "";
 my $dname = "";
 if ($fmt eq "sheet") {
  $r = create_histo_data_file($ba,$var,$oname,$normal);
  if (func_fail($r)) { return $r; };
 } else {
  my $title = '';
  if (defined($a->{options}{title})) {
   $title = $a->{options}{title}{str};
  }
  $dname = $oname . ".tmp.dat";
  $r = create_histo_data_file($ba,$var,$dname,$normal);
  if (func_fail($r)) { return $r; };
  $cname = $oname . ".tmp.cmd";
  $r = create_histo_cmd_file($ba,$var,$fmt,$dname,$cname,
                       $oname,$normal->{do_normal},$title);
  if (func_fail($r)) { return $r; };
 }

 if (1) {
  $oo->disp("Calling gnuplot...");
  my $res = `$self->{gnuplot} $cname`;
  $oo->disp(sprintf("done. %s\n",length($res)?"result: $res":""));
  if (-e $oname) {
   $oo->disp("Created plot \"$oname\"\n");
   $oo->disp(" (bins=$bincount)\n");
  } else {
   return { status => "fail; graph was not created. Is gnuplot installed?" }
  }

  if (!$graph_debug) {
   $oo->disp("");
   unlink $cname;
   unlink $dname;
  }
#  if ($^O eq "MSWin32") {
#   if ($fmt eq $GRAPH_DEFAULT_FORMAT) {
#    if (fork()) {
#    } else {
#     system($oname);
#     exit(0);
#    }
#   }
#  }
 }

 return $r;

}


sub create_histo_data_file {
 my $ba      = shift;
 my $varname = shift;
 my $fname   = shift;
 my $normal  = shift;

 open(OF,"> " . $fname) || return { status=> "fail; could not open temp gnuplot datafile"};
 print OF "$varname\tfrequency\n";
 for (my $b=0;$b<@{$ba};$b++) {
  my $frequency = $ba->[$b]{freq};
  my $center    = $ba->[$b]{center};
  print OF "$center\t$frequency\n";
 }

 if (defined($normal) && $normal->{do_normal}) {
  print OF "\n\n";
  my $min      = $normal->{min};
  my $max      = $normal->{max};
  my $sd       = $normal->{sd};
  my $mean     = $normal->{mean};
  my $obs      = $normal->{obs};
  my $bincount = $normal->{bincount};
  my $points   = $bincount * 10;
  
  if ($points>$MAXPOINTS) { $points = $MAXPOINTS; };
  for (my $x=$min;$x<$max;$x+=($max-$min)/$points) {
   my $z = ($x-$mean)/$sd;
   my $p = (1/sqrt($TWOPI))*exp(-$z*$z/2);
   my $f = $p*($obs/$bincount)*(sqrt($TWOPI))*($PI/2);
   print OF "$x\t$f\n";
  }
 }

 close OF;
 return { status => "ok" };
}

sub create_histo_cmd_file {
 my $ba       = shift;
 my $varname  = shift;
 my $fmt      = shift;
 my $dname    = shift;
 my $cname    = shift;
 my $oname    = shift;
 my $do_norm  = shift;
 my $title    = shift;

 if ($fmt eq "eps") {
  $fmt = "postscript eps enhanced color";
 }

 open (OF,"> " . $cname) || return { status => "fail; could not open temp gnuplot cmd file" };
 my $os = "";
 $os .= "set term $fmt size 2048,1536\n";
 $os .= "set output \"$oname\"\n";
 $os .= "set xlabel \'$varname\'\n";
 $os .= "set ylabel \'frequency\'\n";
 if (defined($title) && length($title)) {
  $os .= "set title \"$title\"\n";
 }
 $os .= "set xtics rotate (";
 for (my $b=0;$b<@{$ba};$b++) {
  $os .= "\"$ba->[$b]{name}\" $ba->[$b]{center},";
 }
 chop $os;
 $os .= ")\n";
 $os .= "plot \"$dname\" index 0 using 1:2 title \'$varname\' with imp lw 5,";
 if ($do_norm) {
  $os .= "     \"$dname\" index 1 using 1:2 title \'normal\' with line\n";
 } else {
  chop $os;
  $os .= "\n";
 }
 $os .= "quit\n";
 print OF $os;
 close OF;
 return { status => "ok" };
};

1;

