
#   Copyright (C) 2006 David G. Jacobowitz 
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#


package matrix_ols;
use Exporter 'import';
@EXPORT_OK = qw(m_ols_regress m_robust_ols m_calc_ols_wald_s m_2sls_regress m_2sls_robust);

use strict;
use matrix_help;
use matlibctrl qw(use_pari use_pdl);
use matrix_help qw(m_gen_tp m_ss_col_s m_return_rank_s 
                   m_sq_elems m_col_to_diag m_mul_m_diag_m
		   m_print_mat_str);

my $using_pari = use_pari();
my $using_pdl  = use_pdl();

if      ($using_pdl) {
 use PDL;
 use pdl_helpers qw(pdl_inv2);
} elsif ($using_pari) {
 use Math::MatrixReal;
} else {
}

use misc qw(deep_copy);
use Data::Dumper;


## simple OLS regression from and IV matrix and DV column vector,
## returning a dispersion matrix and solution vector
sub m_ols_regress {
 my $X = shift;
 my $Y = shift;

 if ($using_pdl) {

  my ($cX,$rX) = $X->dims(); 
  my $tX = $X->transpose();
  my $tXX = $tX x $X;
  # my ($rank,$invertible) = m_return_rank_s($tXX);
  # temporarily skip rank finding because we don't have PDL 
  # function for it. :(
  if (0) {
   return (undef,undef,undef,undef);
  }
  # my $V  = $tXX->inv({s=>1});
  my $V  = pdl_inv2($tXX,{s=>1});
  if (!defined($V)) {
   return (undef,undef,undef,undef);
  }
  my $V_tX    = $V x $tX;
  $tX = undef;
  my $B       = $V_tX x $Y;
  $V_tX = undef;
  my $predict = $X x $B;
  $X = undef;
  my $resid   = $Y - $predict;
  my $rss     = m_ss_col_s($resid,0,1);
  my $msr     = $rss / ($rX-$cX);
  $V         *= $msr;
  return ($V,$B,$predict,$resid,$rss);

 } elsif ($using_pari) {

  my $sz = matsize($X);
  my $rX = $sz->[0];
  my $cX = $sz->[1];

  my $tX = mattranspose($X);
  my $tXX = $tX * $X;
  my ($rank,$invertible) = m_return_rank_s($tXX);
  if (!$invertible) {
   return (undef,undef,undef,undef);
  }

  my $V   = $tXX**-1;
  if (!defined($V)) {
   return (undef,undef,undef,undef);
  }

  # my $V_tX = matsolve($tXX,$tX);
  my $V_tX    = $V * $tX;
  my $B       = $V_tX * $Y;

  my $predict = $X * $B;
  my $resid   = $Y - $predict;

  my $rss     = m_ss_col_s($resid,0,1);
  my $msr     = $rss / ($rX-$cX);
  $V          *= $msr;

  return ($V, $B, $predict, $resid, $rss);
 } else {
  (my $rX, my $cX) = $X->dim;

  my $tX      = Math::MatrixReal->new($cX,$rX);
  $tX->transpose($X);
  my $tXX     = $tX * $X;

  my ($rank,$invertible) = m_return_rank_s($tXX);
  if (!$invertible) {
   return (undef,undef,undef,undef,undef);
  }

  my $V       = $tXX->inverse;
  if (!defined($V)) {
   return (undef,undef,undef,undef,undef);
  }
  my $V_tX    = $V * $tX;
  my $B       = $V_tX * $Y;
  my $predict = $X * $B;
  my $resid   = $Y - $predict;

  my $rss     = m_ss_col_s($resid,0,1);
  my $msr     = $rss / ($rX-$cX);
  $V          *= $msr;
  return ($V, $B, $predict, $resid, $rss);
 }
}

#                  sum( r_hat_ij * u_hat_i_sq)
# Var(Beta_j) =    ----------------------------
#                   SST_j_sq
#
#
# hc0 = inv(t(X)*X)*t(X)*mdiag(resid2)*X*inv(t(X)*X)
# hc1 HC0 * n/(n-k)
#
sub m_robust_ols {
	
 my $X = shift;
 my $Y = shift;
 if ($using_pdl) {
  my ($cols,$rows) = $X->dims();
  my $tX      = $X->transpose();
  my $tXX     = $tX x $X;

#  my ($rank,$invertible) = m_return_rank_s($tXX);
  # Perl::PDL does not have routines to calculate rank. This is 
  # sad as it makes it impossible to detect collinearity.
  # PDL::LinearAlgebra has'em, but it I can't get am easy port 
  # of that for Win32 right now
  if (0) {
   return (undef,undef,undef,undef);
  }
  #my $tXX_inv = $tXX->inv({s=>1});
  my $tXX_inv = pdl_inv2($tXX,{s=>1});
  if (!defined($tXX_inv)) {
   return (undef,undef,undef,undef);
  }
  my $V_tX    = $tXX_inv x $tX;
  my $B       = $V_tX x $Y;
  my $predict = $X x $B;;
  my $resid   = $Y - $predict;
  my $resid2  = $resid * $resid;

  my $rss     = m_ss_col_s($resid,0,1);
  my $HC0_p1  = $tXX_inv x $tX;
  my $HC0_p2  = m_mul_m_diag_m($HC0_p1,$resid2);
  my $HC0     = $HC0_p2 x $X x $tXX_inv;
  my $HC1     = $HC0 * ($rows/($rows-$cols));
  return ($HC1, $B, $predict, $resid, $rss);
 } elsif ($using_pari) {
  my $sz      = matsize($X);
  my $rows    = $sz->[0];
  my $cols    = $sz->[1];
  my $tX      = mattranspose($X);
  my $tXX     = $tX * $X;

  my ($rank,$invertible) = m_return_rank_s($tXX);
  if (!$invertible) {
   return (undef,undef,undef,undef);
  }

  my $tXX_inv = $tXX**-1;
  my $V_tX    = $tXX_inv * $tX;
  my $B       = $V_tX * $Y;
  my $predict = $X * $B;
  my $resid   = $Y - $predict;
  my $resid2  = m_sq_elems($resid);
  my $rss     = m_ss_col_s($resid,0,1);
  my $HC0_p1  = $tXX_inv * $tX;
  my $HC0_p2  = matmuldiagonal($HC0_p1,$resid2->[0]);
  my $HC0     = $HC0_p2 * $X * $tXX_inv;
  my $HC1     = $HC0 * ($rows/($rows-$cols));

  return ($HC1, $B, $predict, $resid, $rss);
 } else {
  (my $rows, my $cols) = $X->dim;
  my $tX      = m_gen_tp($X);
  my $tXX     = $tX*$X;

  my ($rank,$invertible) = m_return_rank_s($tXX);
  if (!$invertible) {
   return (undef,undef,undef,undef,undef);
  }

  my $tXX_inv = $tXX->inverse;
  my $V_tX    = $tXX_inv * $tX;
  my $B       = $V_tX * $Y;
  my $predict = $X * $B;
  my $resid   = $Y - $predict;
  my $rss     = m_ss_col_s($resid,0,1);
  my $resid2  = m_sq_elems($resid);
  my $HC0_p1  = $tXX_inv * $tX;
  my $HC0_p2  = m_mul_m_diag_m($HC0_p1,$resid2);
  my $HC0     = $HC0_p2 * $X * $tXX_inv;
#  my $dr2     = m_col_to_diag($resid2);
#  my $HC0     = $tXX_inv*$tX*$dr2*$X*$tXX_inv;
  my $HC1     = $HC0 * ($rows/($rows-$cols));
  return ($HC1,$B,$predict,$resid,$rss);
 }
}

sub m_dummy_s {
 my $R_m = shift;
 my $r_m = shift;
 my $b_m = shift;
 my $V_m = shift;

 my $bread   = ($R_m * $b_m) - $r_m;
 my $bread_t = mattranspose($bread);
 my $meat    = $R_m * $V_m * mattranspose($R_m);
 my $meat_i  = $meat**-1;
 my $W       = ($bread_t * $meat_i) * $bread;

 my $sz      = matsize($R_m);
 my $R_rows  = $sz->[0];
 my $R_cols  = $sz->[1];
 return ($W->[0][0], $R_rows); 
}

sub m_calc_ols_wald_s {

 my $R_m    = shift;
 my $r_m    = shift;
 my $b_m    = shift;
 my $disp_m = shift;

 my ($R_rows, $R_cols);
 if ($using_pdl) {
  my ($R_cols,$R_rows) = $R_m->dims();
  my $Rb            = $R_m x $b_m;
  my $Rb_r          = $Rb - $r_m;
  my $Rt            = $R_m->transpose();
  my $R_disp_Rt     = $R_m x $disp_m x $Rt;
  my $R_disp_Rt_inv = $R_disp_Rt->inv();
  my $Rb_rt         = $Rb->transpose();
  my $W             = $Rb_rt x $R_disp_Rt_inv x $Rb_r;
  return ($W->at(0,0), $R_rows);
 } elsif ($using_pari) {
  my $Rb            = $R_m * $b_m;
  my $Rb_r          = $Rb - $r_m;
  my $Rt            = mattranspose($R_m);
  my $R_disp_Rt     = $R_m * $disp_m * $Rt;
  my $R_disp_Rt_inv = $R_disp_Rt**-1;
  my $Rb_rt         =  mattranspose($Rb_r);
  my $W             = $Rb_rt * $R_disp_Rt_inv * $Rb_r;
  my $sz            = matsize($R_m);
  $R_rows        = $sz->[0];
  $R_cols        = $sz->[1];
  return (pari2num($W->[0][0]), pari2num($R_rows));
 } else {
  ($R_rows, $R_cols) = $R_m->dim();
  my $step_a = $R_m * $disp_m;
  my $step_b = m_gen_tp($R_m);
  my $step_c = $step_a * $step_b;
  my $step_d = $step_c->inverse;
  my $step_e = $R_m * $b_m;
  my $step_f = Math::MatrixReal->new($step_e->dim());
  $step_f->subtract($step_e,$r_m);
  my ($rs,$cs) = $step_f->dim();
  my $step_g = Math::MatrixReal->new($cs,$rs);
  $step_g->transpose($step_f); 
  my $step_h = $step_g * $step_d;
  my $step_i = $step_h * $step_f;
  my $matrix_stuff = $step_i->element(1,1);
  return ($matrix_stuff, $R_rows);
 }
};

sub m_2sls_regress {
 my $X = shift;
 my $Z = shift;
 my $Y = shift;

 if ($using_pdl) {

  my ($cX,$rX) = $X->dims(); 
  my ($cZ,$rZ) = $Z->dims();

  my $tZ        = $Z->transpose();
  my $tZ_Z      = $tZ x $Z;
  my $tZ_Z_inv  = pdl_inv2($tZ_Z);
  my $Z_tZZi_tZ = $Z x $tZ_Z_inv x $tZ;
  my $Xh        = $Z_tZZi_tZ x $X;

  my $tXh       = $Xh->transpose();
  my $tXhXh     = $tXh x $Xh;
  my $V         = pdl_inv2($tXhXh);
  my $V_tXh     = $V x $tXh;
  my $B2        = $V_tXh x $Y;

  my $predict   = $X x $B2;
  my $resid     = $Y - $predict;
  my $rss       = m_ss_col_s($resid,0,1);
  my $msr       = $rss / ($rX-$cX);
  $V           *= $msr;
  return ($V,$B2,$predict,$resid,$rss);
 } elsif ($using_pari) {
  my $szX = matsize($X);
  my $szZ = matsize($Z);
  my $rX  = $szX->[0];
  my $cX  = $szX->[1];
  my $rZ  = $szZ->[0];
  my $cZ  = $szZ->[1];

  my $tZ   = mattranspose($Z);
  my $tZZ  = $tZ * $Z;
  my ($rank,$invertible) = m_return_rank_s($tZZ);
  if (!$invertible) {
   return (undef,undef,undef,undef);
  } 
  my $tZZi      = $tZZ**-1;
  my $Z_tZZi_tZ = $Z * $tZZi * $tZ;
  my $Xh        = $Z_tZZi_tZ * $X;
  my $tXh       = mattranspose($Xh);
  my $tXhXh     = $tXh * $Xh;
  my $V         = $tXhXh**-1;
  my $V_tXh     = $V * $tXh;
  my $B2        = $V_tXh * $Y;

  my $predict   = $X * $B2;
  my $resid     = $Y - $predict;
  my $rss       = m_ss_col_s($resid,0,1);
  my $msr       = $rss / ($rX-$cX);
  $V           *= $msr;
  return ($V,$B2,$predict,$resid,$rss);
 } else {
  (my $rX, my $cX) = $X->dim;
  (my $rZ, my $cZ) = $Z->dim;

  my $tZ      = Math::MatrixReal->new($cZ,$rZ);
  $tZ->transpose($Z);
  my $tZZ     = $tZ * $Z;

  my ($rank,$invertible) = m_return_rank_s($tZZ);
  if (!$invertible) {
   return (undef,undef,undef,undef,undef);
  }
  
  my $tZZi      = $tZZ->inverse;
  my $Z_tZZi_tZ = $Z * $tZZi * $tZ;
  my $Xh        = $Z_tZZi_tZ * $X;
  my $tXh       = Math::MatrixReal->new($cX,$rX);
  $tXh->transpose($Xh);
  my $tXhXh     = $tXh * $Xh;
  ($rank,$invertible) = m_return_rank_s($tXhXh);
  if (!$invertible) {
   return (undef,undef,undef,undef,undef);
  }
  my $V         = $tXhXh->inverse;
  if (!defined($V)) {
   return (undef,undef,undef,undef,undef);
  }
  my $V_tXh     = $V * $tXh;
  my $B2        = $V_tXh * $Y;

  my $B         = $V_tXh * $Y;
  my $predict   = $X * $B;
  my $resid     = $Y - $predict;

  my $rss       = m_ss_col_s($resid,0,1);
  my $msr       = $rss / ($rX-$cX);
  $V           *= $msr;
  return ($V, $B2, $predict, $resid, $rss);
 }
};

sub m_2sls_robust {
 my $X = shift;
 my $Z = shift;
 my $Y = shift;

 if ($using_pdl) {
  my ($cX,$rX) = $X->dims(); 
  my ($cZ,$rZ) = $Z->dims();

  my $tZ         = $Z->transpose();
  my $tZZ        = $tZ x $Z;
  my $tZZi       = pdl_inv2($tZZ);
  my $Z_tZZi_tZ  = $Z x $tZZi x $tZ;
  my $Xh         = $Z_tZZi_tZ x $X;

  my $tXh        = $Xh->transpose();
  my $tXhXh      = $tXh x $Xh;
  my $tXhXhi     = pdl_inv2($tXhXh,{s=>1});
  if (!defined($tXhXhi)) {
   return (undef,undef,undef,undef);
  };
  my $V_tXh      = $tXhXhi x $tXh;
  my $B2         = $V_tXh x $Y;
  my $predict    = $X x $B2;
  my $resid      = $Y - $predict;
  my $resid2     = $resid * $resid;
  my $rss        = m_ss_col_s($resid,0,1);
  my $HC0_p1     = $tXhXhi x $tXh;
  my $HC0_p2     = m_mul_m_diag_m($HC0_p1,$resid2);
  my $HC0        = $HC0_p2 x $Xh x $tXhXhi;
  my $HC1        = $HC0 * ($rX/($rX-$cX));
  return ($HC1,$B2,$predict,$resid,$rss);
 } elsif ($using_pari) {
   return (undef,undef,undef,undef);
 } else {

  (my $rX, my $cX) = $X->dim;
  (my $rZ, my $cZ) = $Z->dim;

  my $tZ      = m_gen_tp($Z);
  my $tZZ     = $tZ*$Z;

  my ($rank,$invertible) = m_return_rank_s($tZZ);
  if (!$invertible) {
   return (undef,undef,undef,undef,undef);
  } 
  my $tZZi    = $tZZ->inverse;
  my $ZtZZitZ = $Z * $tZZi * $tZ;
  my $Xh      = $ZtZZitZ * $X;

  my $tXh     = m_gen_tp($Xh);
  my $tXhXh   = $tXh*$Xh;

  ($rank,$invertible) = m_return_rank_s($tXhXh);
  if (!$invertible) {
   return (undef,undef,undef,undef,undef);
  }

  my $tXhXhi  = $tXhXh->inverse;
  my $V_tXh   = $tXhXhi * $tXh;
  my $B       = $V_tXh * $Y;
  my $predict = $X * $B;
  my $resid   = $Y - $predict;
  my $rss     = m_ss_col_s($resid,0,1);
  my $resid2  = m_sq_elems($resid);
  my $HC0_p1  = $tXhXhi * $tXh;
  my $HC0_p2  = m_mul_m_diag_m($HC0_p1,$resid2);
  my $HC0     = $HC0_p2 * $Xh * $tXhXhi;
  my $HC1     = $HC0 * ($rX/($rX-$cX));
  return ($HC1,$B,$predict,$resid,$rss);
 }
}
1;
