#!/usr/bin/perl -w

#   Copyright (C) 2006 David G. Jacobowitz 
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#

package parseargs;
use Exporter qw(import);
use POSIX qw(floor);
use eqn_string qw(create_evaluable_eqn);
use misc2 qw(func_fail);
use Data::Dumper;

use Scalar::Util qw(looks_like_number);

@EXPORT_OK = qw(parseArgs);

## this subroutine divides a string into an array by spaced, but 
## preserves quoted substrings as one chunk
sub slow_space_split {
 my $s = shift;
 my $q = "outside";
 my $t = "not_in_word";
 my @r = ();
 my $chunk = "";
 my $lastch = "";
 for (my $i=0;$i<length($s);$i++) {
  my $inch = substr($s,$i,1);
  if ($q eq "inside") {
   if (($inch eq '"')) {
    $q = "outside";
    push(@r,$chunk);
    $chunk = "";
   } else {
    $chunk .= $inch;
   }
  } elsif ($q eq "outside") {
   if ($t eq "not_in_word") {
    if (($inch eq '"')) {
     $q = "inside";
    } elsif ($inch =~ /\S/) {
     $t = "in_word";
     $chunk .= $inch;
    } else {
     $t = "not_in_word";
     $chunk = "";
    }
   } elsif ($t eq "in_word") {
    if (($inch eq '"')) {
     push(@r,$chunk);
     $q = "inside";
     $t = "not_in_word";
     $chunk = "";
    } elsif ($inch =~ /\S/) {
     $chunk .= $inch;
    } else {
     $t = "not_in_word";
     push(@r,$chunk);
     $chunk = "";
    }
   } else {
    # print "ERR - we should never be here, ever\n";
   }
  } else {
   # print "ERR - we should never be here, ever\n";
  }
  $lastch = $inch;
 }
 if (($q eq "inside") || ($t eq "in_word")) {
  push(@r,$chunk);
  $chunk = "";
 }

# foreach $p (@r) {
#  print "PIECE $p\n";
# }
 return @r;
}

## quote string aware split on comma -- only catches
## the first comma!
sub slow_first_comma_split {
 my $s = shift;
 my $q = "outside";
 my $c = 0;
 my $chunk = "";
 my @r = ();
 my $lastch ="";
 for (my $i=0;$i<length($s);$i++) {
  my $inch = substr($s,$i,1);
  if ($q eq "outside") {
   if (($inch eq '"')) {
    $q = "inside";
   } else {
    if (!$c && ($inch eq ',')) {
     push(@r,$chunk);
     $c = 1;
     $chunk = "";
     $inch = ""; # so we don't pick up the comma
    }
   }
  } else {
   if (($inch eq '"')) {
    $q = "outside";
   }
  }
  $chunk .= $inch;
  $lastch = $inch;
 }
 if (length($chunk)) {
  push(@r,$chunk);
 } 
 return @r;
}

sub parseArgs {
 my $inline = shift;
 my $dm     = shift;
 my $gb     = shift;

 my $a      = {};
 
 $a->{cmd}  = "";

 chomp $inline;

 # for debugging; a special command to get the un-masticated command
 # line into the perl interpreter. This is not used for *ANY* at-ats
 # commands, just for development debugging
 if ($inline =~ /^\s*(perl|pdisp)\s*/) {
  my $pline = $inline;
  $pline =~ s/^\s*(perl|pdisp)\s*//;
  $a->{perlline} = $pline;
 }

 $inline = lc($inline);

 # zeroeth: check for any prepended command modifiers:
 if ($inline =~ /^\s*quietly\s*/) {
  $inline =~s/^\s*quietly\s*//;
  $a->{quietly} = 1; 
  $gb->{quietly} = 1;
 }

 # prefix to command: another prepended modifier: "bysort foo:"
 # this one repeats the command by all the tabulate values of foo
 if ($inline =~ /^\s*bysort\s*([a-z]\w*)\s*:/) {
  my $bs = $1;
  $inline =~ s/^\s*bysort\s*([a-z]\w*)\s*://;
  if (defined($bs)) {
   $a->{have_bysort} = 1; 
   $a->{bysortvar}   = $bs;
  } else {
   $a->{have_bysort} = 0; 
  }
 } else {
  $a->{have_bysort} = 0; 
 };

 # first, separate command, arguments and optional flags
 my @flags = ();
 my $before_comma = "";
 my $after_comma = undef; 
 if ($inline =~ /,/) {
  #($before_comma, $after_comma) = split(/,/,$inline);
  ($before_comma, $after_comma) = slow_first_comma_split($inline);
  if (!defined($after_comma)) { $after_comma = ""; }
  $after_comma =~ s/^\s*//;
  $after_comma =~ s/\s*$//;
 } else {
  $before_comma = $inline;
 }
 $before_comma =~ s/^\s*//;
 $before_comma =~ s/\s*$//;

 #(my $cmd, my @args) = split(/\s+/,$before_comma);
 (my $cmd, my @args) = slow_space_split($before_comma);
 if (defined($after_comma)) {
  @flags = split(/\s+/,$after_comma);
 }
 if (!defined($cmd)) { $cmd = ""; }

 if ($before_comma =~ /^\s$|^$|^\/\/|^#|^\*/) {
  $a->{cmd} = "null"; 
  $a->{comment} = $inline;
  return $a;
 }

 $a->{cmd}           = $cmd;
 $a->{args}          = \@args;
 $a->{flgs}          = \@flags;
 $a->{have_by}       = 0;
 $a->{have_gen}      = 0;
 $a->{have_if}       = 0;
 $a->{if_ok}         = 0;
 $a->{have_eval_str} = 0;
 $a->{have_opt}      = 0;

 my $if_str = "";
 my $rest_is_if = 0;
 my $next_is_using = 0;
 my $using = "";
 my $varcount = 0;
 my $next_is_eval_str = 0;
 my $eval_str = "";
 my $next_is_opt = 0;
 my $opt_str  = "";
 my $have_opt = 0;

 # parse out various command clauses including "if", "using", the 
 # LHS and RHS of "=", and [foo=bar] options
 for (my $i=0;$i<@{$a->{args}};$i++) {
#  print "arg : $i : $a->{args}[$i]\n";
  if ($next_is_using) {
   $using .= $a->{args}[$i] . " ";
   $a->{have_using} = 1;
   if ($using =~ /^"/) {
    $next_is_using = 1;
   } else {
    $next_is_using = 0;
   }
   $using =~ s/\s*$//;
   $a->{using} = $using;
  } elsif ($rest_is_if) {
   $if_str .= $a->{args}[$i] . " ";
  } elsif ($next_is_eval_str && ($a->{args}[$i] ne "if")) {
   $a->{have_eval_str} = 1;
   $eval_str .= $a->{args}[$i] . " ";
  } elsif ($next_is_opt) {
   $opt_str .= $a->{args}[$i];
   if ($a->{args}[$i] =~ /\]$/) {
    $next_is_opt = 0;
   }
  } elsif ($a->{args}[$i] eq "if") {
   $rest_is_if = 1; 
   $next_is_eval_str = 0;
  } elsif ($a->{args}[$i] eq "using") {
   $next_is_using = 1; 
# dave was here 4/23
  } elsif ($a->{args}[$i] =~ /^=/) {
   my $targ = $a->{args}[$i];
   $next_is_eval_str = 1;
   $targ =~ s/^=//;
   $eval_str .= $targ . " ";
  } elsif ($a->{args}[$i] =~ /^\[/) {
   $have_opt = 1;
   $a->{have_opt} = 1;
   $opt_str .= $a->{args}[$i];
   if ($a->{args}[$i] =~ /\]$/) {
    $next_is_opt = 0;
   } else {
    $next_is_opt = 1;
   }
  } else {
   my $pvar = $a->{args}[$i];
   $a->{varary}[$varcount]{name}   = $pvar;
   $a->{varary}[$varcount]{state} = "noexist";
   $a->{vars}{$pvar} = "noexist";
   if (defined($dm->{data}{$pvar})) {
    $a->{vars}{$pvar} = "exist";
    $a->{varary}[$varcount]{state} = "exist";
   } elsif (defined($dm->{locals}{$pvar})) {
    $a->{vars}{$pvar} = "local";
    $a->{varary}[$varcount]{state} = "local";
   }
   $varcount++;
  }
 }

 # handle "[foo=bar]" options
 if ($a->{have_opt}) {
  $opt_str =~ s/\[//g;
  my @opts = split(/\]/,$opt_str);
  foreach my $o (@opts) {
   (my $opt, my $eqn) = split(/=/,$o);
   $opt =~ s/\s//g;
   $a->{options}{$opt}{present} = 1;
   if (defined($eqn) && length($eqn)) {
    $a->{options}{$opt}{str} = $eqn;
   }
  }
 }

 # handle and cleanup "using" -- also, implement shorthand
 # file name expansion for "using"
 if ($a->{have_using}) {
  $a->{using} =~ s/^"//;
  $a->{using} =~ s/"$//;
  if (looks_like_number($a->{using})) {
   if (defined($gb->{lastdir}{files}[$a->{using}])) {
    my $repl_using = $gb->{lastdir}{directory} . "/" .
                     $gb->{lastdir}{files}[$a->{using}];
    $a->{using} = $repl_using;
   }
  }
 }

 ## create, for convenience in some functions, ways
 ## to quickly pull off the first variable and the 
 ## rest of the variables;
 $a->{firstvar}{present} = $varcount ? 1 : 0;
 $a->{firstvar}{name}    = $a->{varary}[0]{name};
 $a->{firstvar}{state}   = $a->{varary}[0]{state};

 my @vars = @{$a->{varary}};
 if (@vars>=2) {
  shift @vars;
  $a->{latervars} = \@vars;
 } else {
  $a->{latervars}{present} = 0
 }

 $a->{varcount} = keys %{$a->{vars}};

 # handle the if clause, if present, including evaluating it
 # and generating the 'true' bitmask (create_valuable_eqn does
 # the bitmask generation and places it in r->{ifary}
 if (length($if_str)) {
  $a->{have_if} = 1;
  $a->{if_str}  = $if_str;
  my $tr = create_evaluable_eqn($dm,$gb,$if_str,1);
  if (func_fail($tr)) { $a->{if_ok} = 0; }
  else {
   $a->{ifary}           = $tr->{ifary};
   $a->{reverse_iftable} = $tr->{reverse_iftable};
   $a->{forward_iftable} = $tr->{forward_iftable};
   $a->{if_ok} = 1;
  } 
 }

 # handle the case where an '=' apprears in the command as an
 # assignment. (used in lset and other set commands)
 if (length($eval_str)) {
  $a->{have_eval} = 1;
  $a->{eval_str}      = $eval_str;
 }

 # search all the ',' flags and decode a couple of them that 
 # require a bit more attention...
 foreach my $f (@{$a->{flgs}}) {
  $a->{flag}{$f} = 1;
  if ($f =~ /by\(([a-z]\w*)\)/) {
   # this is used by ttest foo, by(bar)
   # it is probably useful for other commands that will/should/
   # might-one-day accept a by() modifier.
   my $byvar = $1;
   if (defined($byvar)) {
    $a->{have_by} = 1;
    $a->{byvar}   = $byvar;
   }
  } elsif ($f =~ /gen\(([a-z]\w*)\)/) {
   # this is used by tab foo, gen(bar)
   # it's not valid on other circumstances, but we can parse for 
   # it anyway
   my $genvar = $1;
   if (defined($genvar)) {
    $a->{have_gen} = 1;
    $a->{genvar} = $genvar;
   }
  }
 }
 return $a;
};

1;

