#!/usr/bin/perl -w

#    xport.pm - a Perl module to read "TRANSPORT" format files exported from the
#    SAS statistical package, so that they can be manipulated in Perl-land
#
#    Copyright (C) 2008 David G. Jacobowitz

use warnings 'all';
use strict;

package xport;

use Data::Dumper;
use Time::Local qw(timelocal);

=head1 NAME xport

a cheap and simple interface to READ SAS ".xpt" or TRANSPORT files. TRANSPORT 
is an ancient and crufty file format used to export SAS data.

xport.pm can currently *read* such files, but does not have the capability
to write them. I consider this a good thing, as the format is barbaric.


=head1 SYNOPSIS

 use xport;

 my $p = new xport;

 # open a file to be read. Returns 1 on successfull open.
 $p->openFile("mydata.xpt");

 # parse the file. Returns the database if successful, otherwise
 # an empty hash.
 my $db = $p->parse();

 # probably the easiest way to understand what you get back is 
 # to dump it and look.
 print Dumper $db;

 # A convenience feature: dump the database to comma-separated-
 # values files. The name of the file created in this case will
 # be "foo_SASLIB_DATASET.csv", where SASLIB and DATASET are the
 # names of the SAS library and dataset as specified in the .xpt
 # file.
 $p->writeCSV("foo");


=head1 REQUIRES

Time::Local


=head1 EXPORTS

(nothing)


=head1 DESCRIPTION

xport is a simple object-oriented interface for READING "xpt" or TRANSPORT 
files from the statistics program "SAS". Basically, it just reads the
file and returns a hash that looks like this:

 parseReturnValue = 
  {libraries}
   {SAS_LIB_NAME}
    {datasets}
     {DATASETNAME}
      {vars => [ variable descriptors ],
       data => [observations][variables] }

It appears that an XPT file can only include one dataset from one library,
so this tree will usually only have one branch for SAS_LIB_NAME and one
branch for DATASETNAME. However, you re-use the xport object for reading
multiple TRANSPORT files, and the data struct will be added to each time
you call ->parse(), so it is possible to build one struct that contains
multiple separate datasets.

Whether this is useful or not is not immediately apparent.

The "vars" branch points to an array of variable descriptors, each
is itself is a hash of descriptor variables, the most interesting of
which are probably:

  the name   (nname), 
  the label  (nlable),
  the length (nlng), and
  the type   (ntype)

TRANSPORT only has two types; numeric (2) and string (1). All numerics are
stored in floating point format, according to a prehistoric IBM mainframe
floating point format.

The observations branch is an array of observations (duh), each of which
is itself a pointer to an array of data, which are in the same order of the
variable descriptors in the "vars" array.


=head1 AUTHOR

David G. Jacobowitz (david.jacobowitz@gmail.com)


=head1 SEE ALSO

 The SAS website description of the TRANSPORT format:
   http://support.sas.com/techsup/technote/ts140.html#

 A nice Wikipedia description of the IBM floating point format:
   http://en.wikipedia.org/wiki/IBM_Floating_Point_Architecture
   (note that the format used in TRANSPORT is the 64b variety)


=head1 LICENSE

Released freely under the GNU Public License, version 2


=head1 VERSION

Version 0.0:   22 April, 2008


=head1 METHODS

=cut

=head2 xport->new()

Creates a new object of the xport class. It requires a no arguments,
and returns a reference to the object. The class can be re-used for
multiple file parses, if desired.

 my $xp = new xport;


=cut

sub new {
 my $this  = shift;
 my $class = ref($this) || $this;
 my $self  = { valid => 0 };
 bless( $self, $class );
 $self->__initialize();
 return $self;
}

=head2 xport->setFlipEndian()

I wrote and tested this on an x86 PC using .xpt files that were also
generated on such a PC. In this case, it was necessary to flip the
endianness of shorts, longs, and floats, so that is the default behavior
of this code.

If you are running this script on a machine with the opposite endiannes
and it is not working, you may want to try calling:

 $xp->setFlipEndian(0);


=cut

sub setFlipEndian {
 my $s = shift;
 my $n = shift;
 if ($n) {
  $s->{flip_required} = 1;
 } else {
  $s->{flip_required} = 0;
 }
 return 1;
}

=head2 xport->setDetailedMissing()

SAS supports a whole host of different ways to represent missing values,
bad values, not applicables, etc. If you want to see these individuated,
call

 $xp->setDetailedMissing(1)

before you call ->parse(). However, in most cases you just want to 
have emptiness (zero length strings) represent a missing value, and 
that is the default behavior.


=cut

sub setDetailedMissing {
 my $s = shift;
 my $m = shift;
 if ($m) {
  $s->{fancy_missing} = 1;
 } else {
  $s->{fancy_missing} = 0;
 }
}

=head2 xport->openFile($fn)

Open a file to be parsed. Returns true on success, and false
on failure. Do not call $xp->parse() before having opened a file
with this function

 $xp->openFile("health_data.xpt");


=cut

sub openFile {
 my $s  = shift;
 my $fn = shift;

 my $fh = undef;
 if ( -e $fn ) {
  open( $fh, "<", $fn ) or die "Could not open input file $fn\n";
  binmode($fh);
 } else {
  $s->{file_open} = undef;
  push( @{ $s->{status} }, "error: could not open file $fn\n" );
 }
 $s->{fh}        = $fh;
 $s->{file_open} = $fn;
 $s->{endfile}   = 0;
 defined( $s->{file_open} );
}


=head2 xport->setFH($fh)

If you've already opened your file by some other means, and you
just want to hand a file handle to the xport object, use this 
method instead of $xp->openFile(). 

This function, can be used, for example, if you would like to 
directly convert a file that exists on a socket.

 NB 1 :
   Beware, this filehandle *NEEDS* to be set to "binmode" and byte-mode.
   If you don't do this, disaster is certain.

 NB 2 : 
   xport will still CLOSE this file handle for you when it is finished 
   with it or when it is destroyed. This may not be what you want.


=cut

sub setFH {
 my $s  = shift;
 my $fh = shift;
 $s->{file_open} = "something";
 $s->{endfile}   = eof($fh);
 $s->{fh}        = $fh;
 return (defined($s->{file_open}));
}


=head2 xport->parse()

This is the main public function for xport.pm, and it parses the xpt
file previously opened by $xp->openFile(). There's not much to report here.
If all goes well, you will receive as your return value a datastructure
containing the data from your XPORT file. If not, you'll get some piece of 
that datastructure, which may help you understand where xport went wrong.

 $xp->parse() 

closes data files automatically, so you do not have to.


=cut

sub parse {
 my $s = shift;

 my $o                  = $s->{last_parse};
 my $currlib            = '__undef';
 my $currds             = '__undef';
 my $varcount           = 0;
 my $var_descriptor_len = 140;
 my $record_len         = 80;

 $s->__get80();
 while ( !$s->{endfile} ) {
  while ( !$s->{endfile} && ( $s->__headerType() ne "library" ) ) {
   $s->__get80();
  }
  $s->__get80();
  my $r = $s->__parseLibraryHeader();
  if ( $r->{ok} ) {
   $currlib = $r->{saslib};
   $o->{libraries}{$currlib} = {
    date   => $r->{sas_create_tm},
    sasver => $r->{sasver},
    sasos  => $r->{sasos}
   };
   $s->__get80();
   $r = $s->__parseLibraryHeaderL2();
   if ( $r->{ok} ) {
    $o->{libraries}{$currlib}{modified} = $r->{modified_tm};
   }
   $s->__get80();
   while ( $s->__headerType() ne "member" )     { $s->__get80(); }
   while ( $s->__headerType() ne "descriptor" ) { $s->__get80(); }
   $s->__get80();
   $r = $s->__parseDatasetHeader();
   if ( $r->{ok} ) {
    $currds = $r->{sas_dsname};
    $o->{libraries}{$currlib}{datasets}{$currds}{created} = $r->{sas_create_tm};
    $s->__get80();
    $r = $s->__parseDatasetHeaderL2();
    if ( $r->{ok} ) {
     $o->{libraries}{$currlib}{datasets}{$currds}{modified} = $r->{ds_date_tm};
     $o->{libraries}{$currlib}{datasets}{$currds}{label}    = $r->{ds_label};
     $o->{libraries}{$currlib}{datasets}{$currds}{type}     = $r->{ds_type};
     $s->__get80();
     $r = $s->__parseNamstrHeader();
     if ( $r->{ok} ) {
      $varcount = $r->{ds_variable_count};
      $o->{libraries}{$currlib}{datasets}{$currds}{varcount} = $varcount;
      $r = $s->__parseDescriptors($varcount);
      $o->{libraries}{$currlib}{datasets}{$currds}{vars} = $r;
      $s->__get80();
      while ( $s->__headerType() ne "observation" ) { $s->__get80(); }
      $o->{libraries}{$currlib}{datasets}{$currds}{data} = $s->__parseData2($r);
     }
    }
   }
  }
  $s->__get80();
 }

 $s->closeFile();
 return $o;
}

=head2 xport->closeFile()

Close a file previously opened by $xp->openFile().

In the general case, it is not necessary to call this function, as
files are closed automatically after successful parsing. Even if not,
files are closed when the $xp object is DESTROYED. Nevertheless, you
might want to call this if a call to ->parse() has not succeeeded.


=cut

sub closeFile {
 my $s = shift;
 if ( defined( $s->{file_open} ) ) {
  close( $s->{fh} );
  $s->{file_open} = undef;
  $s->{endfile}   = 1;
 }
}

=head2 xport->writeCSV($prefix)

Writes a CSV (comma-separated-values text file for each dataset currently
stored in the $xp object. If you just parsed a single XPORT file, then
you'll get a single .csv file named:

 $prefix . "_" . $LIBNAME . "_" . $DATASETNAME . ".csv"

where $prefix is the lone argument to writeCSV(), $LIBNAME is the
name of the SAS library as specified in the XPORT file, and 
$DATASETNAME is the name of the data set as specified in the XPORT
file.

If you parsed multiple files since creating your xport object, 
a .csv file will be generated for each of the relevant datasets.

Returns a list of the names of the files generated, if any.


=cut

sub writeCSV {
 my $s     = shift;
 my $fn    = shift;
 my @names = ();
 foreach my $lib ( keys %{ $s->{last_parse}{libraries} } ) {
  my $libref = $s->{last_parse}{libraries}{$lib};
  foreach my $ds ( keys %{ $libref->{datasets} } ) {
   my $dsref = $libref->{datasets}{$ds};

   my $fn       = $fn . "_" . $lib . "_" . $ds . ".csv";
   my $os       = '';
   my @varnames = map( ( $_->{nname} =~ /,/ )
    ? '"' . $_->{nname} . '"'
    : $_->{nname},
    @{ $dsref->{vars} } );
   $os .= join( ",", @varnames ) . "\n";
   foreach my $obs ( @{ $dsref->{data} } ) {
    my @oline = map( ( $_ =~ /,/ ) ? '"' . $_ . '"' : $_, @{$obs} );
    $os .= join( ",", @oline ) . "\n";
   }

   open( OF, ">", $fn ) or die "Could not open file $fn for write.\n";
   print OF $os;
   close OF;
   push( @names, $fn );
  }
 }
 return @names;
}

# *** #####################################################
# *** Everything beyond here is private to xport.pm
# *** #####################################################

sub __initialize {
 my $self = shift;
 $self->{status}     = ["ok: __START"];
 $self->{rchunks}    = [ '', '', '', '', '', '', ];
 $self->{last_parse} = {};

 # show missing values by type if true; otherwise just empty
 $self->{fancy_missing} = 0;

 # do we need to flip endianness?
 $self->{flip_required} = 1;
}

sub __getN {
 my $s    = shift;
 my $req  = shift;
 my $lref = shift;
 my $l    = 0;
 my $d    = '';
 if ((defined($s->{file_open})) && (!eof($s->{fh}))) {
  $l = read($s->{fh}, $d, $req);
  if ($l == $req) {
   $s->{rchunks}[1] = $s->{rchunks}[0];
   $s->{rchunks}[0] = $d;
   $s->{endfile}    = 0;
  } else {
   $s->{endfile} = 1;
  }
 } else {
  $s->{endfile} = 1;
 }
 $$lref = $l;
 return $d
}


sub __get80 {
 my $s   = shift;
 my $req = 80;
 my $d   = '';
 if ( ( defined( $s->{file_open} ) ) && ( !eof( $s->{fh} ) ) ) {
  my $l = read( $s->{fh}, $d, $req );
  if ( $l == $req ) {
   $s->{rchunks}[1] = $s->{rchunks}[0];
   $s->{rchunks}[0] = $d;
   $s->{endfile}    = 0;
  } else {
   $s->{endfile} = 1;
  }
 }
 else {
  $s->{endfile} = 1;
 }
 return $d;
}

# remove whitespace around strings
sub __remWhite {
 my $s = shift;
 my $p = $s;
 $p =~ s/^\s*//;
 $p =~ s/\s*$//;
 return $p;
}

sub __headerType {
 my $s    = shift;
 my $d    = $s->{rchunks}[0];
 my $type = "not_a_header";
 if ( $d =~ /^HEADER RECORD/ ) {
  if    ( $d =~ /\*{7}LIBRARY HEADER RECORD!{7}/ ) { $type = "library"; }
  elsif ( $d =~ /\*{7}MEMBER  HEADER RECORD!{7}/ ) { $type = "member"; }
  elsif ( $d =~ /\*{7}DSCRPTR HEADER RECORD!{7}/ ) { $type = "descriptor"; }
  elsif ( $d =~ /\*{7}NAMESTR HEADER RECORD!{7}/ ) { $type = "namestr"; }
  elsif ( $d =~ /\*{7}OBS     HEADER RECORD!{7}/ ) { $type = "observation"; }
  else                                             { $type = "unknown_header"; }
 }
 return $type;
}

sub __parseDateTime {
 my $dt = shift;
 my ( $date, $hour, $minute, $second ) = split( /:/, $dt );
 my $dom = substr( $date, 0, 2 );
 my $mon = lc( substr( $date, 2, 3 ) );
 my $yr = substr( $date, 5, 2 );
 my $mo =
     ( $mon eq 'jan' ) ? 1
   : ( $mon eq 'feb' ) ? 2
   : ( $mon eq 'mar' ) ? 3
   : ( $mon eq 'apr' ) ? 4
   : ( $mon eq 'may' ) ? 5
   : ( $mon eq 'jun' ) ? 6
   : ( $mon eq 'jul' ) ? 7
   : ( $mon eq 'aug' ) ? 8
   : ( $mon eq 'sep' ) ? 9
   : ( $mon eq 'oct' ) ? 10
   : ( $mon eq 'nov' ) ? 11
   : ( $mon eq 'dec' ) ? 12
   :                     -1;
 if   ( $yr < 20 ) { $yr += 2000; }
 else              { $yr += 1900; }

 my $res = timelocal( $second, $minute, $hour, $dom, $mo - 1, $yr );
 return $res;
}

# Capture the first line indicating which library this
# data came from.
sub __parseLibraryHeader {
 my $s = shift;
 my $d = $s->{rchunks}[0];

 my $r = { ok => 0, };
 if ( length($d) ) {
  my $a  = substr( $d, 0,  8 );
  my $b  = substr( $d, 8,  8 );
  my $c  = substr( $d, 16, 8 );
  my $d1 = substr( $d, 24, 8 );
  my $e  = substr( $d, 32, 8 );
  my $f  = substr( $d, 64, 16 );
  $r->{sas_symbol}    = __remWhite( $a . $b );
  $r->{saslib}        = __remWhite($c);
  $r->{sasver}        = __remWhite($d1);
  $r->{sas_os}        = __remWhite($e);
  $r->{sas_create}    = __remWhite($f);
  $r->{sas_create_tm} = __parseDateTime($f);
  $r->{ok}            = 1;
 }
 return $r;
}

# The library header is followed by a line containing the last
# modify time for the library.
sub __parseLibraryHeaderL2 {
 my $s = shift;
 my $d = $s->{rchunks}[0];
 my $r = { ok => 0, };

 if ( length($d) ) {
  my $dt = substr( $d, 0, 16 );
  my $dtm = __parseDateTime($dt);
  $r->{modified}    = __remWhite($dt);
  $r->{modified_tm} = $dtm;
  $r->{ok}          = 1;
 }
 return $r;
}

# Now, parse the header that indicates what dataset within
# the library is to follow
sub __parseDatasetHeader {
 my $s = shift;
 my $d = $s->{rchunks}[0];
 my $r = { ok => 0, };
 if ( length($d) ) {
  my $sas_sym     = substr( $d, 0,  8 );
  my $sas_dsname  = substr( $d, 8,  8 );
  my $sas_sasdata = substr( $d, 16, 8 );
  my $sas_sasver  = substr( $d, 24, 8 );
  my $sas_sasos   = substr( $d, 32, 8 );
  my $sas_create  = substr( $d, 64, 16 );
  my $sas_create_tm = __parseDateTime($sas_create);
  $r = {
   sas_sym       => __remWhite($sas_sym),
   sas_dsname    => __remWhite($sas_dsname),
   sasdata       => __remWhite($sas_sasdata),
   sasver        => __remWhite($sas_sasver),
   sas_osname    => __remWhite($sas_sasos),
   sas_create    => $sas_create,
   sas_create_tm => $sas_create_tm,
   ok            => 1,
  };
 }
 return $r;
}

# the dataset header is a two-line structure, the second
# part of which contains the modified date and a label
sub __parseDatasetHeaderL2 {
 my $s = shift;
 my $d = $s->{rchunks}[0];
 my $r = { ok => 0 };
 if ( length($d) ) {
  my $modified      = substr( $d, 0,  16 );
  my $dataset_label = substr( $d, 32, 40 );
  my $dataset_type  = substr( $d, 71, 8 );
  $r = {
   ds_label   => __remWhite($dataset_label),
   ds_type    => __remWhite($dataset_type),
   ds_date    => $modified,
   ds_date_tm => __parseDateTime($modified),
   ok         => 1,
  };
 }
 return $r;
}

# When we get to the NAMESTR header, we find out how many
# variables are in the dataset. WE should check to make
# sure this conforms to the header format but we don't. It
# may not be necessary.
sub __parseNamstrHeader {
 my $s = shift;
 my $d = $s->{rchunks}[0];
 my $r = { ok => 0, };

 if ( length($d) ) {
  my $ds_var_count = substr( $d, 54, 4 );
  $r = {
   ds_variable_count => $ds_var_count,
   ok                => 1,
  };
 }
 return $r;
}

# reverse endianness for a short
sub __flipEndian16 {
 my $d = shift;
 my $f = shift;
 if ($f) {
  return (
   ( ( ( $d & 0xff ) << 8 ) & 0xff00 ) | 
   ( ( ( $d & 0xff00 ) >> 8 ) & 0xff ) 
  );
 } else {
  return $d;
 }
}

# reverse endianness for a long
sub __flipEndian32 {
 my $d = shift;
 my $f = shift;
 if ($f) {
  return ( 
     ( ( ( $d & 0xff ) << 24 ) & 0xff000000 ) |
     ( ( ( $d & 0xff00 ) <<  8 ) & 0xff0000 ) |
     ( ( ( $d & 0xff0000 ) >>  8 ) & 0xff00 ) |
     ( ( ( $d & 0xff000000 ) >> 24 ) & 0xff ) 
  );
 } else {
  return $d;
 }
}

# Continue to parse the NAMESTR records. This routine unpacks
# all the information for each variable.
sub __parseNameStr {
 my $d = shift;
 my $f = shift;

 my $ps = "SSSSA8A40A8SSSA2A8SSLA52";
 my @p  = unpack( $ps, $d );
 my $r  = {
  ntype  => __flipEndian16( $p[0],  $f ),
  nhfun  => __flipEndian16( $p[1],  $f ),
  nlng   => __flipEndian16( $p[2],  $f ),
  nvar0  => __flipEndian16( $p[3],  $f ),
  nname  => __remWhite( $p[4] ),
  nlabel => __remWhite( $p[5] ),
  nform  => __remWhite( $p[6] ),
  nfl    => __flipEndian16( $p[7],  $f ),
  nfd    => __flipEndian16( $p[8],  $f ),
  nfj    => __flipEndian16( $p[9],  $f ),
  nfill  => __remWhite( $p[10] ),
  niform => __remWhite( $p[11] ),
  nifl   => __flipEndian16( $p[12], $f ),
  nifd   => __flipEndian16( $p[13], $f ),
  npos   => __flipEndian32( $p[14], $f ),
  ok     => 1,
 };
 return $r;
}

# parseNameStrHeader can understand a single variable descriptor.
# This routine determines how many there are, how many block of
# 80 bytes that will require, then loads the blocks together
# and parses them.
sub __parseDescriptors {
 my $s              = shift;
 my $c              = shift;
 my $rec_len        = 80;
 my $descriptor_len = 140;

 # get the descriptor data as a big string
 my $d = '';
 my $recs_required = int( ( $descriptor_len * $c / $rec_len ) + 0.5 );
 for my $i ( 1 .. $recs_required ) {
  $d .= $s->__get80();
 }

 # parse it out and store it.
 my $o = [];
 for my $i ( 0 .. $c - 1 ) {
  my $di = substr( $d, $i * $descriptor_len, $descriptor_len );
  my $r = __parseNameStr( $di, $s->{flip_required} );
  my $name = $r->{nname};
  $o->[$i] = $r;
 }
 return $o;
}

# helper function determines, in bytes, the length of an observation
# in the dataset.
sub __calcObsLength {
 my $varspecs = shift;
 my $l        = 0;
 my $vc       = @$varspecs;
 for my $i ( 0 .. $vc - 1 ) {
  $l += $varspecs->[$i]{nlng};
 }
 return ( $l, $vc );
}

sub __parseData2 {
 my $s        = shift;
 my $varspecs = shift;

 my ($olength, $vcount) = __calcObsLength($varspecs);

 my $d = ''; my $rlength = 0;
 my $data = [];
 my $ct   = 0;

 while (!$s->{endfile}) {
  $d = $s->__getN($olength,\$rlength);
  if ($rlength == $olength) {
   my $obs = [];
   for my $j ( 0 .. $vcount-1) {
    my $varw = $varspecs->[$j]{nlng};
    my $varp = $varspecs->[$j]{npos};
    my $val  = substr($d,$varp,$varw);
    if ($varspecs->[$j]{ntype} == 1) {
     my @char8 = unpack( "CCCCCCCC", $val );
     $val = __IBMtoNative(
      {
       data          => \@char8,
       flip          => $s->{flip_required},
       fancy_missing => $s->{fancy_missing}
      }
     );
    } else {
     $val = __remWhite($val);
    }
    $obs->[$j] = $val;
   }
   $data->[$ct++] = $obs;
  }
 }
 return $data;
}


=cut

This is the old version, which was probably faster but required
the entire XPORT data section to be loaded in memory. Probably not 
very good.

sub __parseData {
 my $s        = shift;
 my $varspecs = shift;

 # first, read in the remaining data
 # ... maybe I'll modify this not to read in all the data at once, so
 # as not to completely suck up memory needlessly.
 my $d      = '';
 my $rcount = 0;
 while ( !$s->{endfile} ) {
  my $l = $s->__get80();
  if ( length($l) == 80 ) {
   $d .= $l;
   $rcount++;
  }
 }

 my $data = [];
 my $dl   = length($d);
 my ( $obslength, $varcount ) = __calcObsLength($varspecs);
 my $obs = int( $dl / $obslength );

 for my $i ( 0 .. $obs - 1 ) {
  my $start_index = $obslength * $i;
  for my $j ( 0 .. $varcount - 1 ) {
   my $varw   = $varspecs->[$j]{nlng};
   my $obsloc = $varspecs->[$j]{npos};
   my $val    = substr( $d, $start_index + $obsloc, $varw );
   if ( $varspecs->[$j]{ntype} == 1 ) {
    my @char8 = unpack( "CCCCCCCC", $val );
    $val = __IBMtoNative(
     {
      data          => \@char8,
      flip          => $s->{flip_required},
      fancy_missing => $s->{fancy_missing}
     }
    );
   } else {
    $val = __remWhite($val);
   }
   $data->[$i][$j] = $val;
  }
 }
 return $data;
}
=cut

# convert an FP value from IBM 64b format to something reasonable
# the SAS people should be ashamed of themselves for this bullsh*
sub __IBMtoNative {
 my $q     = shift;
 my @chars = @{ $q->{data} };
 my $f     = $q->{flip};
 my $m     = $q->{fancy_missing};

 my $v = undef;

 if ($f) {
  my $t;
  $t        = $chars[7]; $chars[7] = $chars[0]; $chars[0] = $t;
  $t        = $chars[6]; $chars[6] = $chars[1]; $chars[1] = $t;
  $t        = $chars[5]; $chars[5] = $chars[2]; $chars[2] = $t;
  $t        = $chars[4]; $chars[4] = $chars[3]; $chars[3] = $t;
 }

 if ( $chars[6] | $chars[5] | $chars[4] | $chars[3] | 
      $chars[2] | $chars[1] | $chars[0] ) {

  my $sign = ( $chars[7] & 0x80 ) ? 1 : 0;
  my $exp  = ( $chars[7] & 0x7f );

  my $fraction_l =
    ( ( $chars[0] & 0xff ) ) | 
    ( ( $chars[1] <<  8 ) & 0xff00 ) |
    ( ( $chars[2] << 16 ) & 0xff0000 ) | 
    ( ( $chars[3] << 24 ) & 0xff000000 );

  my $fraction_h =
    ( ( $chars[4] & 0xff ) ) | 
    ( ( $chars[5] <<  8 ) & 0xff00 ) |
    ( ( $chars[6] << 16 ) & 0xff0000 );

  my $fraction = ( $fraction_h * ( 2**32 ) + $fraction_l ) / ( 2**56 );

  $v = $fraction * ( 16**( $exp - 64 ) );

  if ($sign) { $v = -$v; }

  if (0) {
   print sprintf(
    "%x %x %x %x %x %x %x %x\n",
    $chars[7], $chars[6], $chars[5], $chars[4],
    $chars[3], $chars[2], $chars[1], $chars[0]
   );
   print "sign $sign exp $exp fraction $fraction value $v\n";
  }

 } else {

  # handle missing values
  $v = $chars[7] ? $m ? "__missing_" . chr( $chars[7] ) : '' : 0;
 }
 return $v;
}

# for cleanliness at exit
sub DESTROY {
 my $s = shift;
 $s->closeFile();
}

1;

